# -*- coding: utf-8 -*-

import os
import json
import configparser
from typing import List

import xgboost as xgb
import time
import faiss
import requests
import numpy as np
import pandas as pd

from ydl_ai_recommender.src.core.indexer import (
    UserCounselorChatIndexer,
    UserCounselorOrderIndexer,
    UserCounselorDefaultIndexer,
    UserCounselorCombinationIndexer,
    CounselorCounselorCFIndexer,
)
from ydl_ai_recommender.src.core.profile import encode_user_profile
from ydl_ai_recommender.src.utils import get_conf_path, get_data_path, read_user_encoder_dict, read_counselors, get_project_path
from ydl_ai_recommender.src.utils.log import create_logger

logger = create_logger(__name__, 'service_xgb.log', is_rotating=True)

class Recommender():

    def __init__(self) -> None:
        self.logger = create_logger(__name__, 'recommender.log')
        self.default_indexer = UserCounselorDefaultIndexer(self.logger)
        self.default_indexer.load_index_data()

    def _recommend_top(self, size=0):
        return [{
            'counselor': str(c_id),
            'score': score,
            'from': 'default',
        } for [c_id, score] in self.default_indexer.index(size)]

    def recommend(self, user, size=100) -> List:
        raise self._recommend_top(size)


class UserCFRecommender(Recommender):

    def __init__(self, top_n=5, k=20, u2c='combination', c2c=None) -> None:
        """
        params:
            top_n: 每个召回的用户获取的相关咨询师个数
            k: 召回的相似用户数
            u2c: [用户->咨询师] 索引方法
            c2c: [咨询师->咨询师] 索引方法，None 表示不使用咨询师拓展
        """
        super().__init__()
        select_items = [
            'uid', 'country_code', 'channel_id_type', 'ffrom_login', 'user_preference_cate', 'consult_pay_money',
            'listen_pay_money', 'test_items_pay_money', 'course_pay_money', 'consult_order_num',
            'listen_order_num', 'test_items_order_num', 'course_order_num', 'aidi_cst_bias_city',
            'aidi_cst_bias_sex', 'aidi_cst_bias_price', 'aidi_cst_bias_server_type', 'user_login_city',
            'd30_inquire_order_num', 'd30_session_num',
        ]
        self.select_fields = {k: True for k in select_items}
        # 召回 top_n 个相似用户
        self.top_n = top_n
        # 每个召回的用户取 k 个相关咨询师
        self.k = k

        self.logger = create_logger(__name__, 'recommender.log')
        
        config = configparser.RawConfigParser()
        config.read(get_conf_path())
        self.dmp_url = config.get('DMP', 'url')

        if u2c == 'chat':
            self.indexer = UserCounselorChatIndexer(self.logger)
        elif u2c == 'order':
            self.indexer = UserCounselorOrderIndexer(self.logger)
        else:
            self.indexer = UserCounselorCombinationIndexer(self.logger)
        self.indexer.load_index_data()

        if c2c:
            self.c2c_indexer = CounselorCounselorCFIndexer(self.logger)
            self.c2c_indexer.load_index_data()
        else:
            self.c2c_indexer = None

        self.local_file_dir = get_data_path()
        self.load_data()


    def load_data(self):
        order_user_embedding = []
        order_user_ids = []

        with open(os.path.join(self.local_file_dir, 'user_embeddings_ids.txt'), 'r', encoding='utf-8') as f:
            order_user_ids = [line.strip() for line in f]

        with open(os.path.join(self.local_file_dir, 'user_embeddings.json'), 'r', encoding='utf-8') as f:
            order_user_embedding = json.load(f)

        self.order_user_embedding = order_user_embedding
        self.order_user_ids = order_user_ids

        self.index = faiss.IndexFlatL2(len(self.order_user_embedding[0]))
        self.index.add(np.array(self.order_user_embedding))


    def get_user_profile(self, user_id):
        if user_id == '0':
            return []
            
        headers = {
            'X-App-Id': 'plough_cloud',            
            'Content-Type': 'application/json'
        }

        payload = {
            "filter": {
                "uid": user_id,
            },
            "fields": self.select_fields,
            "limit": 10
        }

        try:
            response = requests.request('POST', self.dmp_url, headers=headers, json=payload)
            resp = response.json()
            return resp['data']['objects'][0]
        except Exception as e:
            self.logger.error('获取用户画像数据失败: %s', e, exc_info=True)
            try:
                self.logger.exception('response json data %s', resp)
            except:
                pass
            return []


    def _recommend(self, user_embedding):
        D, I = self.index.search(np.array([user_embedding]), self.k)
        counselors = []

        for idx, simi_score in zip(I[0], D[0]):
            # 相似用户uid
            similar_user_id = self.order_user_ids[idx]
            similar_user_counselor = self.indexer.index(q=similar_user_id, count=self.top_n)
            recommend_data = [{
                'counselor': c_id,
                'score': score / max(0.01, float(simi_score)),
                'from': 'similar_users {}'.format(similar_user_id),
            } for (c_id, score) in similar_user_counselor]

            supplement_data = []
            if self.c2c_indexer:
                for ro in recommend_data:
                    supplement_data.extend([{
                        'counselor': sc_id,
                        'score': ro['score'] * score,
                        # 'score': (ro['score'] + score) / 2,
                        'from': 'supplement {} {}'.format(ro['counselor'], ro['from']),
                    } for (sc_id, score) in self.c2c_indexer.index(ro['counselor'], count=int(self.top_n))])
                    # } for (sc_id, score) in self.c2c_indexer.index(ro['counselor'], count=int(self.top_n / len(recommend_data)))])

            counselors.extend(recommend_data + supplement_data)

        counselors.sort(key=lambda x: x['score'], reverse=True)
        return counselors


    def recommend_with_profile(self, user_profile, size=0, is_merge=True):
        user_embedding = encode_user_profile(user_profile)
        counselors = self._recommend(user_embedding)

        # size == 0 时，不追加默认推荐咨询师
        if size > 0:
            counselors.extend(self._recommend_top())

        if is_merge:
            counselor_set = set()
            merged_counselors = []
            for counselor in counselors:
                if counselor['counselor'] not in counselor_set:
                    counselor_set.add(counselor['counselor'])
                    merged_counselors.append(counselor)
            counselors = merged_counselors

        if size > 0:
            counselors = counselors[:size]
        return counselors


    def recommend(self, user_id, size=0, is_merge=True):
        """
        根据用户画像，推荐咨询师
        若获取不到用户画像，推荐默认咨询师（订单最多的）
        """
        user_profile = self.get_user_profile(user_id)
        if not user_profile:
            return self._recommend_top(size)

        return self.recommend_with_profile(user_profile, size, is_merge)


class ItemCFRecommender(Recommender):

    def __init__(self, top_n=10) -> None:
        super().__init__()
        self.top_n = top_n
        self.logger = create_logger(__name__, 'item_cf_recommender.log')

        self.indexer = UserCounselorChatIndexer(self.logger)
        self.indexer.load_index_data()

        self.c2c_indexer = CounselorCounselorCFIndexer(self.logger)
        self.c2c_indexer.load_index_data()


    def recommend(self, user_id, size=0) -> List:
        similar_user_counselor = self.indexer.index(q=user_id)

        if not similar_user_counselor:
            return self._recommend_top(size)

        recommend_data = [{
            'counselor': c_id,
            'score': score * 100,
            'from': 'similar_users {}'.format(user_id),
        } for (c_id, score) in similar_user_counselor]

        supplement_data = []
        for ro in recommend_data:
            supplement_data.extend([{
                'counselor': sc_id,
                'score': ro['score'] * score,
                'from': 'supplement {} {}'.format(ro['counselor'], ro['from']),
            } for (sc_id, score) in self.c2c_indexer.index(ro['counselor'], count=int(self.top_n))])

        counselors = recommend_data + supplement_data + self._recommend_top()
        counselors.sort(key=lambda x: x['score'], reverse=True)

        if size > 0:
            counselors = counselors[:size]
        return counselors

class RecommendByXgboost(Recommender):
    def __init__(self, top_n=5, k=20, is_use_db=True, u2c='combination', c2c=None) -> None:
        super().__init__()

        config = configparser.RawConfigParser()
        config.read(get_conf_path())
        self.dmp_url = config.get('DMP', 'url')

        select_items = ['uid', 'ffrom_login', 'user_login_city', 'user_preference_cate']
        self.select_fields = {k: True for k in select_items}
        self.user_encoder_convert = read_user_encoder_dict()

        self.all_counselors = read_counselors()

        #self.recommender = UserCFRecommender(top_n=top_n, k=k, u2c=u2c)
        self.params = {'n_estimators': 150, 'max_depth': 7, 'min_child_weight': 5, 'gamma': 0, 'subsample': 0.9,
                       'colsample_bytree': 0.5, 'reg_alpha': 0, 'reg_lambda': 1, 'learning_rate': 0.1,
                       'max_delta_step': 0,
                       'scale_pos_weight': 1}
        self.model = xgb.XGBClassifier(objective='binary:logistic', nthread=-1, **self.params)
        self.model.load_model(os.path.join(get_project_path(), 'model_data/xgb_model.bin'))

    def recommend(self, user_id, size=0, is_merge=True):

        s_u_profile_time = time.time()
        user_profile = self.get_user_profile(user_id)
        logger.info('s_u_profile_time: {} '.format(time.time()-s_u_profile_time))

        if not user_profile:
            return self._recommend_top(size)
        recall_start = time.time()
        # recommend_result = self.recommender.recommend(user_id, size=size, is_merge=True)
        recommend_result = None
        logger.info('recall call time:{}'.format(time.time()-recall_start))

        data_time = time.time()
        predit_data = self.trans_feature_data(user_id, recommend_result)
        logger.info('data_time: {}'.format(time.time()-data_time))
        doctor_ids = predit_data.pop('doctor_id')
        doctor_ids = doctor_ids.to_numpy()

        pre_time = time.time()
        predit_result = self.model.predict_proba(predit_data)[:, 1]
        logger.info('predit_time:{}'.format(time.time()-pre_time))
        result_dict = dict(zip(doctor_ids, predit_result))
        result_dict = sorted(result_dict.items(), key=lambda x:x[1], reverse=True)
        recommend_data = [{
            'counselor': int(c_id),
            'score': float(proba),
            'from': 'similar_users {}'.format(user_id),
        } for (c_id, proba) in result_dict[0:50]]
        return recommend_data

    def trans_feature_data(self, user_id, counselor_data):
        user_feature_data = self.trans_user_feature_data(user_id)
        counselor_feature_data = self.trans_counselor_feature_data(counselor_data)
        counselor_num = len(counselor_feature_data)

        user_feature_data_dataframe = pd.DataFrame([user_feature_data]*counselor_num, columns=['ffrom_login_encoder'\
            , 'user_login_city_encoder', 'cate_id_1_encoder', 'cate_id_2_encoder', 'cate_id_3_encoder'\
            , 'cate_id_4_encoder', 'cate_id_5_encoder'])

        predit_feature_data = pd.concat([user_feature_data_dataframe, counselor_feature_data], axis=1)
        return predit_feature_data

    def trans_user_feature_data(self, user_id):
        user_profile = self.get_user_profile(user_id)

        from_login_encoder = self.get_encoder_from_dict('ffrom_login', user_profile['ffrom_login'])
        user_login_city_encoder = self.get_encoder_from_dict('user_login_city', user_profile['user_login_city'])

        user_preference_cate = user_profile['user_preference_cate']
        user_preference_cate_top_5_encoder = self.process_user_preference_cate(user_preference_cate)

        user_feature_data = [from_login_encoder, user_login_city_encoder]
        user_feature_data.extend(user_preference_cate_top_5_encoder)

        return user_feature_data

    def trans_counselor_feature_data(self, counselor_data):
        # counselor_ids = [str(item['counselor']) for item in counselor_data]
        # counselor_profiles = self.all_counselors[self.all_counselors['doctor_id'].isin(counselor_ids)].reset_index(drop=True)
        # return counselor_profiles
        return self.all_counselors

    def get_user_profile(self, user_id):
        if user_id == '0':
            return []

        headers = {
            'X-App-Id': 'plough_cloud',
            'Content-Type': 'application/json'
        }

        payload = {
            "filter": {
                "uid": user_id,
            },
            "fields": self.select_fields,
            "limit": 10
        }

        try:
            get_profile_time = time.time()
            response = requests.request('POST', self.dmp_url, headers=headers, json=payload)
            self.logger.info(' get user profile cost {} ms'.format((time.time()-get_profile_time)*1000))
            resp = response.json()
            return resp['data']['objects'][0]
        except Exception as e:
            self.logger.error('获取用户画像数据失败: %s', e, exc_info=True)
            try:
                self.logger.exception('response json data %s', resp)
            except:
                pass
            return []

    def process_user_preference_cate(self, preference_cate):
        result = [0, 0, 0, 0, 0]
        ids = []
        if isinstance(preference_cate, str):
            pref_data = json.loads(preference_cate)
            for info in pref_data:
                ids.append(info['cate_id'])
            ids = ids[0:min(5, len(ids))]
        for ind, val in enumerate(ids):
            result[ind] = val
        encoder_result = []
        for ind, val in enumerate(result):
            value_convert_dict = self.user_encoder_convert.get('cate_id_{}_encoder'.format(ind+1))
            if value_convert_dict is not None:
                encoder_result.append(value_convert_dict.get(val, 0))
        if len(encoder_result)<5:
            encoder_result.extend([0]*(5-len(encoder_result)))


        return encoder_result

    def get_encoder_from_dict(self, feature_name, feature_value):
        value_convert_dict = self.user_encoder_convert.get('{}_encoder'.format(feature_name))
        if value_convert_dict is None:
            return 0
        return value_convert_dict.get(str(feature_value), 0)
    def aa(self):
        pass


if __name__ == '__main__':
    s_time = time.time()
    recommender1 = UserCFRecommender()
    recommender1.recommend('30004410')
    print('all cost time: {}'.format(time.time() - s_time), recommender1.recommend('12047'))

    print()
    print()
    s_time = time.time()
    recommender = RecommendByXgboost()
    recommender.recommend('30004410')
    print('all cost time: '.format(time.time()-s_time), recommender.recommend('12047'))

    print()
    print()
    s_time = time.time()
    recommender.recommend('30004410')
    print('all cost time: '.format(time.time() - s_time), recommender.recommend('12047'))